from dataclasses import dataclass
from enum import Enum


class StepUpMfaGrantType(str, Enum):
    ONE_TIME_USE = "ONE_TIME_USE"
    TIME_BASED = "TIME_BASED"


@dataclass
class StepUpMfaVerifyTotpResponse:
    step_up_grant: str

    def __getitem__(self, key):
        return getattr(self, key)
    
@dataclass
class SendSmsMfaCodeResponse:
    challenge_id: str

    def __getitem__(self, key):
        return getattr(self, key)
    
@dataclass
class VerifySmsChallengeResponse:
    step_up_grant: str

    def __getitem__(self, key):
        return getattr(self, key)




from datetime import timedelta

import pytest

from propelauth_py import UnauthorizedException
from propelauth_py.errors import ForbiddenException
from tests.auth_helpers import create_access_token, orgs_to_org_id_map, random_org, random_user_id, random_org_id
from tests.conftest import WRONG_TEST_BASE_AUTH_URL, generate_rsa_keys


def test_validate_without_auth(auth, rsa_keys):
    required_org_id = random_org_id()

    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user_with_org(None, required_org_id)


def test_validate_without_auth_2(auth, rsa_keys):
    required_org_id = random_org_id()

    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user_with_org("", required_org_id)


def test_validate_org_member_with_auth_but_no_org_membership(auth, rsa_keys):
    required_org_id = random_org_id()
    user_id = random_user_id()
    access_token = create_access_token({"user_id": user_id}, rsa_keys.private_pem)

    with pytest.raises(ForbiddenException):
        auth.validate_access_token_and_get_user_with_org("Bearer " + access_token, required_org_id)


def test_validate_org_member_with_auth_and_org_member(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Owner")
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    user_and_org = auth.validate_access_token_and_get_user_with_org("Bearer " + access_token, org["org_id"])

    assert user_and_org.user.user_id == user_id
    assert user_and_org.org_member_info.org_id == org["org_id"]
    assert user_and_org.org_member_info.org_name == org["org_name"]
    assert user_and_org.org_member_info.user_assigned_role == "Owner"


def test_validate_org_member_with_auth_but_wrong_org_id(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Owner")
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    # Pass wrong org_id as required
    wrong_org_id = random_org_id()
    with pytest.raises(ForbiddenException):
        auth.validate_access_token_and_get_user_with_org("Bearer " + access_token, wrong_org_id)


def test_validate_org_member_with_auth_but_no_permission(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Member")
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    # Require at least admin, but the user is a member
    with pytest.raises(ForbiddenException):
        auth.validate_access_token_and_get_user_with_org_by_minimum_role("Bearer " + access_token, org["org_id"], "Admin")


def test_validate_org_member_with_auth_with_permission(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin")
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    user_and_org = auth.validate_access_token_and_get_user_with_org_by_minimum_role("Bearer " + access_token, org["org_id"], "Admin")

    assert user_and_org.user.user_id == user_id
    assert user_and_org.org_member_info.org_id == org["org_id"]
    assert user_and_org.org_member_info.org_name == org["org_name"]
    assert user_and_org.org_member_info.user_assigned_role == "Admin"


def test_validate_org_member_with_auth_by_permissions(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin", ["permA"])
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    user_and_org = auth.validate_access_token_and_get_user_with_org_by_permission("Bearer " + access_token, org["org_id"], "permA")

    assert user_and_org.user.user_id == user_id
    assert user_and_org.org_member_info.org_id == org["org_id"]
    assert user_and_org.org_member_info.org_name == org["org_name"]
    assert user_and_org.org_member_info.user_assigned_role == "Admin"


def test_validate_org_member_by_missing_permission(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin")
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    with pytest.raises(ForbiddenException):
        auth.validate_access_token_and_get_user_with_org_by_permission("Bearer " + access_token, org["org_id"], "permission")


def test_validate_org_member_with_auth_by_batch_permissions(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin", ["permA", "permB"])
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    user_and_org = auth.validate_access_token_and_get_user_with_org_by_all_permissions("Bearer " + access_token, org["org_id"], ["permA"])
    assert user_and_org.user.user_id == user_id
    assert user_and_org.org_member_info.org_id == org["org_id"]
    assert user_and_org.org_member_info.org_name == org["org_name"]
    assert user_and_org.org_member_info.user_assigned_role == "Admin"

    user_and_org = auth.validate_access_token_and_get_user_with_org_by_all_permissions("Bearer " + access_token, org["org_id"], ["permA", "permB"])
    assert user_and_org.user.user_id == user_id
    assert user_and_org.org_member_info.org_id == org["org_id"]
    assert user_and_org.org_member_info.org_name == org["org_name"]
    assert user_and_org.org_member_info.user_assigned_role == "Admin"

    with pytest.raises(ForbiddenException):
        auth.validate_access_token_and_get_user_with_org_by_all_permissions("Bearer " + access_token, org["org_id"], ["permA", "missing"])


def test_validate_org_member_with_bad_header(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin")
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user_with_org("token " + access_token, org["org_id"])


def test_validate_org_member_with_wrong_token(auth, rsa_keys):
    required_org_id = random_org_id()

    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user_with_org("Bearer whatisthis", required_org_id)


def test_validate_org_member_with_expired_token(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Owner")
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem, expires_in=timedelta(minutes=-1))

    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user_with_org("Bearer" + access_token, org["org_id"])


def test_validate_org_member_with_bad_issuer(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Owner")
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem, issuer=WRONG_TEST_BASE_AUTH_URL)

    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user_with_org("Bearer" + access_token, org["org_id"])


def test_validate_org_member_with_wrong_key(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Owner")
    org_id_to_org_member_info = orgs_to_org_id_map([org])

    incorrect_rsa_keys = generate_rsa_keys()
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, incorrect_rsa_keys.private_pem, issuer=WRONG_TEST_BASE_AUTH_URL)

    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user_with_org("Bearer " + access_token, org["org_id"])



from typing import Optional
import requests
import httpx
from propelauth_py.api import _ApiKeyAuth, _is_valid_id, _auth_hostname_header, BACKEND_API_BASE_URL, _get_async_headers
from propelauth_py.errors import RateLimitedException
from dataclasses import dataclass

ENDPOINT_URL = f"{BACKEND_API_BASE_URL}/api/backend/v1/employee"

    
@dataclass
class FetchEmployeeResponse:
    email: str

    def __getitem__(self, key):
        return getattr(self, key)

def _fetch_employee_by_id(auth_hostname, integration_api_key, employee_id) -> Optional[FetchEmployeeResponse]:
    if not _is_valid_id(employee_id):
        return None

    url = f"{ENDPOINT_URL}/{employee_id}"

    response = requests.get(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return None
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching employee")

    json_response = response.json()
    return FetchEmployeeResponse(
        email=json_response.get('email'),
    )
    
async def _fetch_employee_by_id_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    employee_id
) -> Optional[FetchEmployeeResponse]:
    if not _is_valid_id(employee_id):
        return None

    url = f"{ENDPOINT_URL}/{employee_id}"

    response = await httpx_client.get(
        url=url,
        headers=_get_async_headers(auth_hostname=auth_hostname, integration_api_key=integration_api_key)
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return None
    response.raise_for_status()
    json_response = response.json()
    return FetchEmployeeResponse(
       email=json_response.get('email'),
    )


import requests
import json
import httpx
from propelauth_py.api import _ApiKeyAuth, _auth_hostname_header, BACKEND_API_BASE_URL, _get_async_headers
from propelauth_py.errors import (
    BadRequestException,
    RateLimitedException,
    FeatureGatedException,
)
from propelauth_py.types.step_up_mfa import VerifySmsChallengeResponse

ENDPOINT_URL = f"{BACKEND_API_BASE_URL}/api/backend/v1/mfa/step-up/phone/verify"


def _verify_sms_challenge(
    auth_hostname,
    integration_api_key,
    challenge_id,
    user_id,
    code,
) -> VerifySmsChallengeResponse:
    url = ENDPOINT_URL
    json = {
        "challenge_id": challenge_id,
        "user_id": user_id,
        "code": code,
    }

    response = requests.post(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.ok:
        json_response = response.json()
        return VerifySmsChallengeResponse(
            step_up_grant=json_response.get("step_up_grant")
        )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)

    try:
        error_response = response.json()
        error_code = error_response.get("error_code")
        if error_code == "invalid_request_fields":
            raise BadRequestException(error_response.get("user_facing_errors", {}))
        elif error_code == "feature_gated":
            raise FeatureGatedException()
        else:
            raise RuntimeError(
                f"Unknown error when verifying sms challenge: {error_response}"
            )
    except requests.exceptions.JSONDecodeError:
        raise RuntimeError("Unknown error when verifying sms challenge")
    
    
async def _verify_sms_challenge_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    challenge_id,
    user_id,
    code,
) -> VerifySmsChallengeResponse:
    url = ENDPOINT_URL
    json_body = {
        "challenge_id": challenge_id,
        "user_id": user_id,
        "code": code,
    }


    response = await httpx_client.post(
        url=url,
        json=json_body,
        headers=_get_async_headers(auth_hostname=auth_hostname, integration_api_key=integration_api_key),
    )

    if response.is_success:
        json_response = response.json()
        return VerifySmsChallengeResponse(
            step_up_grant=json_response.get("step_up_grant")
        )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)

    try:
        error_response = response.json()
        error_code = error_response.get("error_code")
        if error_code == "invalid_request_fields":
            raise BadRequestException(error_response.get("user_facing_errors", {}))
        elif error_code == "feature_gated":
            raise FeatureGatedException()
        else:
            raise RuntimeError(
                f"Unknown error when verifying sms challenge: {error_response}"
            )
    except json.JSONDecodeError:
        raise RuntimeError("Unknown error when verifying sms challenge")



from uuid import uuid4

from propelauth_py.types.login_method import (
    PasswordLoginMethod,
    SamlLoginProvider,
    SamlSsoLoginMethod,
    SocialLoginProvider,
    SocialSsoLoginMethod,
    UnknownLoginMethod,
)
from propelauth_py.user import _to_user, User, OrgMemberInfo


def test_to_user_without_orgs():
    user_id = str(uuid4())
    email = str(uuid4())
    login_method = {
        "login_method": "password",
    }
    user = _to_user({"user_id": user_id, "email": email, "login_method": login_method})
    expected_user = User(user_id, None, email, login_method=PasswordLoginMethod())
    assert user == expected_user


def test_to_user_with_no_login_method():
    user_id = str(uuid4())
    email = str(uuid4())
    user = _to_user({"user_id": user_id, "email": email})
    expected_user = User(user_id, None, email, login_method=UnknownLoginMethod())
    assert user == expected_user


def test_to_user():
    user_id = str(uuid4())
    email = str(uuid4())
    first_name = str(uuid4())
    last_name = str(uuid4())
    username = str(uuid4())
    org_role_structure = "single_role_in_hierarchy"
    org_a = {
        "org_id": str(uuid4()),
        "org_name": "orgA",
        "org_metadata": {"org_metadata_key_a": "org_metadata_value_a"},
        "user_role": "Owner",
        "inherited_user_roles_plus_current_role": ["Owner", "Admin", "Member"],
        "user_permissions": ["View", "Edit", "Delete"],
        "org_role_structure": org_role_structure,
        "url_safe_org_name": "orga",
        "additional_roles": [],
        "legacy_org_id": "legacy_org_id_a",
    }
    org_b = {
        "org_id": str(uuid4()),
        "org_name": "orgB",
        "org_metadata": {"org_metadata_key_b": "org_metadata_value_b"},
        "user_role": "Admin",
        "inherited_user_roles_plus_current_role": ["Admin", "Member"],
        "user_permissions": ["View", "Edit"],
        "org_role_structure": org_role_structure,
        "url_safe_org_name": "orgb",
        "additional_roles": [],
        "legacy_org_id": "legacy_org_id_b",
    }
    org_c = {
        "org_id": str(uuid4()),
        "org_name": "orgC",
        "org_metadata": {"org_metadata_key_c": "org_metadata_value_c"},
        "user_role": "Member",
        "inherited_user_roles_plus_current_role": ["Member"],
        "user_permissions": ["View"],
        "org_role_structure": org_role_structure,
        "url_safe_org_name": "orgc",
        "additional_roles": [],
    }
    org_id_to_org_member_info = {
        org_a["org_id"]: org_a,
        org_b["org_id"]: org_b,
        org_c["org_id"]: org_c,
    }
    login_method = {
        "login_method": "saml_sso",
        "provider": "Okta",
        "org_id": org_a["org_id"],
    }

    user = _to_user(
        {
            "user_id": user_id,
            "org_id_to_org_member_info": org_id_to_org_member_info,
            "email": email,
            "first_name": first_name,
            "last_name": last_name,
            "username": username,
            "login_method": login_method,
        }
    )

    expected_org_id_to_org_member_info = {
        org_a["org_id"]: OrgMemberInfo(
            org_id=org_a["org_id"],
            org_name=org_a["org_name"],
            org_metadata=org_a["org_metadata"],
            user_assigned_role="Owner",
            user_inherited_roles_plus_current_role=["Owner", "Admin", "Member"],
            user_permissions=["View", "Edit", "Delete"],
            org_role_structure=org_role_structure,
            url_safe_org_name=org_a["url_safe_org_name"],
            assigned_additional_roles=[],
            legacy_org_id="legacy_org_id_a",
        ),
        org_b["org_id"]: OrgMemberInfo(
            org_id=org_b["org_id"],
            org_name=org_b["org_name"],
            org_metadata=org_b["org_metadata"],
            user_assigned_role="Admin",
            user_inherited_roles_plus_current_role=["Admin", "Member"],
            user_permissions=["View", "Edit"],
            org_role_structure=org_role_structure,
            url_safe_org_name=org_b["url_safe_org_name"],
            assigned_additional_roles=[],
            legacy_org_id="legacy_org_id_b",
        ),
        org_c["org_id"]: OrgMemberInfo(
            org_id=org_c["org_id"],
            org_name=org_c["org_name"],
            org_metadata=org_c["org_metadata"],
            user_assigned_role="Member",
            user_inherited_roles_plus_current_role=["Member"],
            user_permissions=["View"],
            org_role_structure=org_role_structure,
            url_safe_org_name=org_c["url_safe_org_name"],
            assigned_additional_roles=[],
            legacy_org_id=None,
        ),
    }
    expected_user = User(
        user_id=user_id,
        org_id_to_org_member_info=expected_org_id_to_org_member_info,
        email=email,
        first_name=first_name,
        last_name=last_name,
        username=username,
        login_method=SamlSsoLoginMethod(
            provider=SamlLoginProvider.OKTA, org_id=org_a["org_id"]
        ),
    )

    assert user == expected_user


def test_to_user_with_active_org():
    user_id = str(uuid4())
    email = str(uuid4())
    first_name = str(uuid4())
    last_name = str(uuid4())
    username = str(uuid4())
    org_role_structure = "single_role_in_hierarchy"
    org_member_info = {
        "org_id": str(uuid4()),
        "org_name": "orgA",
        "org_metadata": {"org_metadata_key_a": "org_metadata_value_a"},
        "user_role": "Owner",
        "inherited_user_roles_plus_current_role": ["Owner", "Admin", "Member"],
        "user_permissions": ["View", "Edit", "Delete"],
        "org_role_structure": org_role_structure,
        "url_safe_org_name": "orga",
        "additional_roles": [],
        "legacy_org_id": "legacy_org_id",
    }
    login_method = {
        "login_method": "social_sso",
        "provider": "Google",
        "org_id": None,
    }

    user = _to_user(
        {
            "user_id": user_id,
            "org_member_info": org_member_info,
            "email": email,
            "first_name": first_name,
            "last_name": last_name,
            "username": username,
            "login_method": login_method,
        }
    )

    expected_org_id_to_org_member_info = {
        org_member_info["org_id"]: OrgMemberInfo(
            org_id=org_member_info["org_id"],
            org_name=org_member_info["org_name"],
            org_metadata=org_member_info["org_metadata"],
            user_assigned_role="Owner",
            user_inherited_roles_plus_current_role=["Owner", "Admin", "Member"],
            user_permissions=["View", "Edit", "Delete"],
            org_role_structure=org_role_structure,
            url_safe_org_name=org_member_info["url_safe_org_name"],
            assigned_additional_roles=[],
            legacy_org_id="legacy_org_id",
        ),
    }
    expected_user = User(
        user_id=user_id,
        org_id_to_org_member_info=expected_org_id_to_org_member_info,
        email=email,
        first_name=first_name,
        last_name=last_name,
        username=username,
        active_org_id=org_member_info["org_id"],
        login_method=SocialSsoLoginMethod(provider=SocialLoginProvider.GOOGLE),
    )

    assert user == expected_user


def test_to_user_multi_role():
    user_id = str(uuid4())
    email = str(uuid4())
    first_name = str(uuid4())
    last_name = str(uuid4())
    username = str(uuid4())
    org_role_structure = "multi_role"
    org_a = {
        "org_id": str(uuid4()),
        "org_name": "orgA",
        "org_metadata": {"org_metadata_key_a": "org_metadata_value_a"},
        "user_role": "Owner",
        "inherited_user_roles_plus_current_role": ["Owner"],
        "user_permissions": ["View", "Edit", "Delete"],
        "org_role_structure": org_role_structure,
        "url_safe_org_name": "orga",
        "additional_roles": ["Billing", "Finance"],
    }
    org_b = {
        "org_id": str(uuid4()),
        "org_name": "orgB",
        "org_metadata": {"org_metadata_key_b": "org_metadata_value_b"},
        "user_role": "Member",
        "inherited_user_roles_plus_current_role": ["Member"],
        "user_permissions": ["View", "Edit"],
        "org_role_structure": org_role_structure,
        "url_safe_org_name": "orgb",
        "additional_roles": ["IT"],
    }
    org_c = {
        "org_id": str(uuid4()),
        "org_name": "orgC",
        "org_metadata": {"org_metadata_key_c": "org_metadata_value_c"},
        "user_role": "Contractor",
        "inherited_user_roles_plus_current_role": ["Contractor"],
        "user_permissions": ["View"],
        "org_role_structure": org_role_structure,
        "url_safe_org_name": "orgc",
        "additional_roles": [],
    }
    org_id_to_org_member_info = {
        org_a["org_id"]: org_a,
        org_b["org_id"]: org_b,
        org_c["org_id"]: org_c,
    }

    user = _to_user(
        {
            "user_id": user_id,
            "org_id_to_org_member_info": org_id_to_org_member_info,
            "email": email,
            "first_name": first_name,
            "last_name": last_name,
            "username": username,
        }
    )

    expected_org_id_to_org_member_info = {
        org_a["org_id"]: OrgMemberInfo(
            org_id=org_a["org_id"],
            org_name=org_a["org_name"],
            org_metadata=org_a["org_metadata"],
            user_assigned_role="Owner",
            user_inherited_roles_plus_current_role=["Owner"],
            user_permissions=["View", "Edit", "Delete"],
            org_role_structure=org_role_structure,
            url_safe_org_name=org_a["url_safe_org_name"],
            assigned_additional_roles=["Billing", "Finance"],
        ),
        org_b["org_id"]: OrgMemberInfo(
            org_id=org_b["org_id"],
            org_name=org_b["org_name"],
            org_metadata=org_b["org_metadata"],
            user_assigned_role="Member",
            user_inherited_roles_plus_current_role=["Member"],
            user_permissions=["View", "Edit"],
            org_role_structure=org_role_structure,
            url_safe_org_name=org_b["url_safe_org_name"],
            assigned_additional_roles=["IT"],
        ),
        org_c["org_id"]: OrgMemberInfo(
            org_id=org_c["org_id"],
            org_name=org_c["org_name"],
            org_metadata=org_c["org_metadata"],
            user_assigned_role="Contractor",
            user_inherited_roles_plus_current_role=["Contractor"],
            user_permissions=["View"],
            org_role_structure=org_role_structure,
            url_safe_org_name=org_c["url_safe_org_name"],
            assigned_additional_roles=[],
        ),
    }
    expected_user = User(
        user_id=user_id,
        org_id_to_org_member_info=expected_org_id_to_org_member_info,
        email=email,
        first_name=first_name,
        last_name=last_name,
        username=username,
        login_method=UnknownLoginMethod(),
    )

    assert user == expected_user



from datetime import timedelta

import pytest

from propelauth_py import UnauthorizedException
from tests.auth_helpers import create_access_token, random_user_id
from tests.conftest import WRONG_TEST_BASE_AUTH_URL, generate_rsa_keys


def test_validate_without_header(auth, rsa_keys):
    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user(None)


def test_validate_with_header(auth, rsa_keys):
    user_id = random_user_id()
    access_token = create_access_token({"user_id": user_id}, rsa_keys.private_pem)
    user = auth.validate_access_token_and_get_user("Bearer " + access_token)
    assert user.user_id == user_id
    assert user.org_id_to_org_member_info is None


def test_validate_with_invalid_header(auth, rsa_keys):
    user_id = random_user_id()
    access_token = create_access_token({"user_id": user_id}, rsa_keys.private_pem)
    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user("token " + access_token)


def test_validate_with_wrong_token(auth, rsa_keys):
    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user("Bearer whatisthis")


def test_validate_with_expired_token(auth, rsa_keys):
    user_id = random_user_id()
    access_token = create_access_token({"user_id": user_id}, rsa_keys.private_pem, expires_in=timedelta(minutes=-5))
    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user("Bearer " + access_token)


def test_validate_with_bad_issuer(auth, rsa_keys):
    user_id = random_user_id()
    access_token = create_access_token({"user_id": user_id}, rsa_keys.private_pem, issuer=WRONG_TEST_BASE_AUTH_URL)
    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user("Bearer " + access_token)


def test_validate_with_wrong_key(auth, rsa_keys):
    incorrect_rsa_keys = generate_rsa_keys()

    user_id = random_user_id()
    access_token = create_access_token({"user_id": user_id}, incorrect_rsa_keys.private_pem)
    with pytest.raises(UnauthorizedException):
        auth.validate_access_token_and_get_user("Bearer " + access_token)



from typing import Any, Optional, Dict, Union
from propelauth_py.user import _to_org_member_info
import requests
import httpx

from propelauth_py.api import (
    _ApiKeyAuth,
    _format_params,
    _is_valid_id,
    BACKEND_API_BASE_URL,
    _auth_hostname_header,
    _get_async_headers,
)
from propelauth_py.api.end_user_api_keys import _validate_api_key, _validate_api_key_async
from propelauth_py.types.user import (
    MfaPhoneType,
    MfaPhones,
    MfaTotpType,
    UserMetadata,
    UsersPagedResponse,
    CreatedUser,
    PersonalApiKeyValidation,
    UserSignupQueryParams,
    FetchUserMfaMethodsResponse
)
from propelauth_py.errors import (
    BadRequestException,
    CreateUserException,
    EndUserApiKeyException,
    InviteUserToOrgException,
    UpdateUserEmailException,
    UpdateUserMetadataException,
    UpdateUserPasswordException,
    RateLimitedException,
)

ENDPOINT_URL = f"{BACKEND_API_BASE_URL}/api/backend/v1/user"


####################
#       GET        #
####################
def _fetch_user_metadata_by_user_id(
    auth_hostname, integration_api_key, user_id, include_orgs=False
) -> Optional[UserMetadata]:
    if not _is_valid_id(user_id):
        return None

    user_info_url = f"{ENDPOINT_URL}/{user_id}"
    query = {"include_orgs": include_orgs}
    return _fetch_user_metadata_by_query(
        integration_api_key, user_info_url, query, auth_hostname
    )
    
async def _fetch_user_metadata_by_user_id_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id,
    include_orgs=False
) -> Optional[UserMetadata]:
    if not _is_valid_id(user_id):
        return None

    user_info_url = f"{ENDPOINT_URL}/{user_id}"
    query = {"include_orgs": include_orgs}
    return await _fetch_user_metadata_by_query_async(
        httpx_client, integration_api_key, user_info_url, query, auth_hostname
    )


def _fetch_user_signup_query_params_by_user_id(
    auth_hostname,
    integration_api_key,
    user_id,
) -> Optional[UserSignupQueryParams]:
    if not _is_valid_id(user_id):
        return None

    user_signup_query_params_url = f"{ENDPOINT_URL}/{user_id}/signup_query_parameters"
    response = requests.get(
        url=user_signup_query_params_url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return None
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching user signup query params")

    json_response = response.json()
    return UserSignupQueryParams(
        user_signup_query_parameters=json_response.get("user_signup_query_parameters")
    )
    
async def _fetch_user_signup_query_params_by_user_id_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id,
) -> Optional[UserSignupQueryParams]:
    if not _is_valid_id(user_id):
        return None

    url = f"{ENDPOINT_URL}/{user_id}/signup_query_parameters"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.get(
        url=url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return None

    response.raise_for_status()

    json_response = response.json()
    return UserSignupQueryParams(
        user_signup_query_parameters=json_response.get("user_signup_query_parameters")
    )


def _fetch_user_metadata_by_email(
    auth_hostname, integration_api_key, email, include_orgs=False
) -> Optional[UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/email"
    query = {"include_orgs": include_orgs, "email": email}
    return _fetch_user_metadata_by_query(
        integration_api_key, user_info_url, query, auth_hostname
    )
    
async def _fetch_user_metadata_by_email_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    email,
    include_orgs=False
) -> Optional[UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/email"
    query = {"include_orgs": include_orgs, "email": email}
    return await _fetch_user_metadata_by_query_async(
        httpx_client, integration_api_key, user_info_url, query, auth_hostname
    )


def _fetch_user_metadata_by_username(
    auth_hostname, integration_api_key, username, include_orgs=False
) -> Optional[UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/username"
    query = {"include_orgs": include_orgs, "username": username}
    return _fetch_user_metadata_by_query(
        integration_api_key, user_info_url, query, auth_hostname
    )
    
async def _fetch_user_metadata_by_username_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    username,
    include_orgs=False
) -> Optional[UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/username"
    query = {"include_orgs": include_orgs, "username": username}
    return await _fetch_user_metadata_by_query_async(
        httpx_client, integration_api_key, user_info_url, query, auth_hostname
    )


def _fetch_user_metadata_by_query(
    integration_api_key, user_info_url, query, auth_hostname
) -> Optional[UserMetadata]:
    response = requests.get(
        user_info_url,
        params=_format_params(query),
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif response.status_code == 404:
        return None
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching user metadata")

    json_response = response.json()

    return UserMetadata(
        user_id=json_response.get("user_id"),
        email=json_response.get("email"),
        email_confirmed=json_response.get("email_confirmed"),
        has_password=json_response.get("has_password"),
        username=json_response.get("username"),
        first_name=json_response.get("first_name"),
        last_name=json_response.get("last_name"),
        picture_url=json_response.get("picture_url"),
        locked=json_response.get("locked"),
        enabled=json_response.get("enabled"),
        mfa_enabled=json_response.get("mfa_enabled"),
        can_create_orgs=json_response.get("can_create_orgs"),
        created_at=json_response.get("created_at"),
        last_active_at=json_response.get("last_active_at"),
        org_id_to_org_info=_to_org_member_info(json_response.get("org_id_to_org_info")),
        legacy_user_id=json_response.get("legacy_user_id"),
        impersonator_user_id=json_response.get("impersonator_user_id"),
        metadata=json_response.get("metadata"),
        properties=json_response.get("properties"),
    )
    
async def _fetch_user_metadata_by_query_async(
    httpx_client: httpx.AsyncClient,
    integration_api_key,
    user_info_url,
    query,
    auth_hostname,
) -> Optional[UserMetadata]:
    headers = _get_async_headers(auth_hostname, integration_api_key)
    params = _format_params(query)

    response = await httpx_client.get(
        url=user_info_url,
        params=params,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise BadRequestException(response.text)
    elif response.status_code == 404:
        return None

    response.raise_for_status()

    json_response = response.json()
        
    return UserMetadata(
        user_id=json_response.get("user_id"),
        email=json_response.get("email"),
        email_confirmed=json_response.get("email_confirmed"),
        has_password=json_response.get("has_password"),
        username=json_response.get("username"),
        first_name=json_response.get("first_name"),
        last_name=json_response.get("last_name"),
        picture_url=json_response.get("picture_url"),
        locked=json_response.get("locked"),
        enabled=json_response.get("enabled"),
        mfa_enabled=json_response.get("mfa_enabled"),
        can_create_orgs=json_response.get("can_create_orgs"),
        created_at=json_response.get("created_at"),
        last_active_at=json_response.get("last_active_at"),
        org_id_to_org_info=_to_org_member_info(json_response.get("org_id_to_org_info")),
        legacy_user_id=json_response.get("legacy_user_id"),
        impersonator_user_id=json_response.get("impersonator_user_id"),
        metadata=json_response.get("metadata"),
        properties=json_response.get("properties"),
    )


def _fetch_batch_user_metadata_by_user_ids(
    auth_hostname, integration_api_key, user_ids, include_orgs
) -> Dict[str, UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/user_ids"
    params = {"include_orgs": include_orgs}
    body = {"user_ids": user_ids}
    return _fetch_batch_user_metadata_by_query(
        user_info_url,
        integration_api_key,
        params,
        body,
        lambda x: x["user_id"],
        auth_hostname,
    )
    
async def _fetch_batch_user_metadata_by_user_ids_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_ids,
    include_orgs,
) -> Dict[str, UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/user_ids"
    params = {"include_orgs": include_orgs}
    body = {"user_ids": user_ids}
    return await _fetch_batch_user_metadata_by_query_async(
        httpx_client,
        user_info_url,
        integration_api_key,
        params,
        body,
        lambda x: x["user_id"],
        auth_hostname,
    )


def _fetch_batch_user_metadata_by_emails(
    auth_hostname, integration_api_key, emails, include_orgs
) -> Dict[str, UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/emails"
    params = {"include_orgs": include_orgs}
    body = {"emails": emails}
    return _fetch_batch_user_metadata_by_query(
        user_info_url,
        integration_api_key,
        params,
        body,
        lambda x: x["email"],
        auth_hostname,
    )
    
async def _fetch_batch_user_metadata_by_emails_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    emails,
    include_orgs,
) -> Dict[str, UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/emails"
    params = {"include_orgs": include_orgs}
    body = {"emails": emails}
    return await _fetch_batch_user_metadata_by_query_async(
        httpx_client,
        user_info_url,
        integration_api_key,
        params,
        body,
        lambda x: x["email"],
        auth_hostname,
    )


def _fetch_batch_user_metadata_by_usernames(
    auth_hostname, integration_api_key, usernames, include_orgs
) -> Dict[str, UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/usernames"
    params = {"include_orgs": include_orgs}
    body = {"usernames": usernames}
    return _fetch_batch_user_metadata_by_query(
        user_info_url,
        integration_api_key,
        params,
        body,
        lambda x: x["username"],
        auth_hostname,
    )
    
async def _fetch_batch_user_metadata_by_usernames_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    usernames,
    include_orgs,
) -> Dict[str, UserMetadata]:
    user_info_url = f"{ENDPOINT_URL}/usernames"
    params = {"include_orgs": include_orgs}
    body = {"usernames": usernames}
    return await _fetch_batch_user_metadata_by_query_async(
        httpx_client,
        user_info_url,
        integration_api_key,
        params,
        body,
        lambda x: x["username"],
        auth_hostname,
    )


def _fetch_batch_user_metadata_by_query(
    user_info_url, integration_api_key, params, body, key_fn, auth_hostname
) -> Dict[str, UserMetadata]:
    response = requests.post(
        user_info_url,
        params=_format_params(params),
        json=body,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching batch user metadata")

    json_response = response.json()
    return_value = {}
    for single_item in json_response:
        return_value[key_fn(single_item)] = single_item

    return return_value

async def _fetch_batch_user_metadata_by_query_async(
    httpx_client: httpx.AsyncClient,
    user_info_url,
    integration_api_key,
    params,
    body,
    key_fn,
    auth_hostname
) -> Dict[str, UserMetadata]:
    headers = _get_async_headers(auth_hostname, integration_api_key)
    formatted_params = _format_params(params)

    response = await httpx_client.post(
        user_info_url,
        params=formatted_params,
        json=body,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise BadRequestException(response.text)

    response.raise_for_status()

    json_response = response.json()
    return_value = {}
    for single_item in json_response:
        return_value[key_fn(single_item)] = single_item

    return return_value


def _fetch_users_by_query(
    auth_hostname,
    integration_api_key,
    page_size,
    page_number,
    order_by,
    email_or_username,
    include_orgs,
    legacy_user_id,
) -> UsersPagedResponse:
    url = f"{ENDPOINT_URL}/query"
    params = {
        "page_size": page_size,
        "page_number": page_number,
        "order_by": order_by,
        "email_or_username": email_or_username,
        "include_orgs": include_orgs,
        "legacy_user_id": legacy_user_id,
    }
    response = requests.get(
        url,
        params=_format_params(params),
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching orgs by query")

    json_response = response.json()

    users = [
        UserMetadata(
            user_id=key.get("user_id"),
            email=key.get("email"),
            email_confirmed=key.get("email_confirmed"),
            has_password=key.get("has_password"),
            username=key.get("username"),
            first_name=key.get("first_name"),
            last_name=key.get("last_name"),
            picture_url=key.get("picture_url"),
            locked=key.get("locked"),
            enabled=key.get("enabled"),
            mfa_enabled=key.get("mfa_enabled"),
            can_create_orgs=key.get("can_create_orgs"),
            created_at=key.get("created_at"),
            last_active_at=key.get("last_active_at"),
            org_id_to_org_info=_to_org_member_info(key.get("org_id_to_org_info")),
            legacy_user_id=key.get("legacy_user_id"),
            impersonator_user_id=key.get("impersonator_user_id"),
            metadata=key.get("metadata"),
            properties=key.get("properties"),
        )
        for key in json_response.get("users")
    ]

    return UsersPagedResponse(
        users=users,
        total_users=json_response.get("total_users"),
        current_page=json_response.get("current_page"),
        page_size=json_response.get("page_size"),
        has_more_results=json_response.get("has_more_results"),
    )
    
async def _fetch_users_by_query_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    page_size,
    page_number,
    order_by,
    email_or_username,
    include_orgs,
    legacy_user_id,
) -> UsersPagedResponse:
    url = f"{ENDPOINT_URL}/query"
    order_by_value = order_by.value if hasattr(order_by, 'value') else order_by
    params = {
        "page_size": page_size,
        "page_number": page_number,
        "order_by": order_by_value,
        "email_or_username": email_or_username,
        "include_orgs": include_orgs,
        "legacy_user_id": legacy_user_id,
    }
    headers = _get_async_headers(auth_hostname, integration_api_key)
    formatted_params = _format_params(params)

    response = await httpx_client.get(
        url,
        params=formatted_params,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise BadRequestException(response.text)
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )

    response.raise_for_status()

    json_response = response.json()

    users = [
        UserMetadata(
            user_id=key.get("user_id"),
            email=key.get("email"),
            email_confirmed=key.get("email_confirmed"),
            has_password=key.get("has_password"),
            username=key.get("username"),
            first_name=key.get("first_name"),
            last_name=key.get("last_name"),
            picture_url=key.get("picture_url"),
            locked=key.get("locked"),
            enabled=key.get("enabled"),
            mfa_enabled=key.get("mfa_enabled"),
            can_create_orgs=key.get("can_create_orgs"),
            created_at=key.get("created_at"),
            last_active_at=key.get("last_active_at"),
            org_id_to_org_info=_to_org_member_info(key.get("org_id_to_org_info")),
            legacy_user_id=key.get("legacy_user_id"),
            impersonator_user_id=key.get("impersonator_user_id"),
            metadata=key.get("metadata"),
            properties=key.get("properties"),
        )
        for key in json_response.get("users")
    ]

    return UsersPagedResponse(
        users=users,
        total_users=json_response.get("total_users"),
        current_page=json_response.get("current_page"),
        page_size=json_response.get("page_size"),
        has_more_results=json_response.get("has_more_results"),
    )


def _fetch_users_in_org(
    auth_hostname,
    integration_api_key,
    org_id,
    page_size,
    page_number,
    include_orgs,
    role,
) -> UsersPagedResponse:
    if not _is_valid_id(org_id):
        return UsersPagedResponse(
            users=[],
            total_users=0,
            current_page=page_number,
            page_size=page_size,
            has_more_results=False,
        )

    url = f"{ENDPOINT_URL}/org/{org_id}"
    params = {
        "page_size": page_size,
        "page_number": page_number,
        "include_orgs": include_orgs,
        "role": role,
    }
    response = requests.get(
        url,
        params=_format_params(params),
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching users in org")

    json_response = response.json()

    users = [
        UserMetadata(
            user_id=key.get("user_id"),
            email=key.get("email"),
            email_confirmed=key.get("email_confirmed"),
            has_password=key.get("has_password"),
            username=key.get("username"),
            first_name=key.get("first_name"),
            last_name=key.get("last_name"),
            picture_url=key.get("picture_url"),
            locked=key.get("locked"),
            enabled=key.get("enabled"),
            mfa_enabled=key.get("mfa_enabled"),
            can_create_orgs=key.get("can_create_orgs"),
            created_at=key.get("created_at"),
            last_active_at=key.get("last_active_at"),
            org_id_to_org_info=_to_org_member_info(key.get("org_id_to_org_info")),
            legacy_user_id=key.get("legacy_user_id"),
            impersonator_user_id=key.get("impersonator_user_id"),
            metadata=key.get("metadata"),
            properties=key.get("properties"),
        )
        for key in json_response.get("users")
    ]

    return UsersPagedResponse(
        users=users,
        total_users=json_response.get("total_users"),
        current_page=json_response.get("current_page"),
        page_size=json_response.get("page_size"),
        has_more_results=json_response.get("has_more_results"),
    )
    
async def _fetch_users_in_org_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    org_id,
    page_size,
    page_number,
    include_orgs,
    role,
) -> UsersPagedResponse:
    if not _is_valid_id(org_id):
        return UsersPagedResponse(
            users=[],
            total_users=0,
            current_page=page_number,
            page_size=page_size,
            has_more_results=False,
        )

    url = f"{ENDPOINT_URL}/org/{org_id}"
    params = {
        "page_size": page_size,
        "page_number": page_number,
        "include_orgs": include_orgs,
        "role": role,
    }
    headers = _get_async_headers(auth_hostname, integration_api_key)
    formatted_params = _format_params(params)

    response = await httpx_client.get(
        url,
        params=formatted_params,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise BadRequestException(response.text)
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )

    response.raise_for_status()

    json_response = response.json()

    users = [
        UserMetadata(
            user_id=key.get("user_id"),
            email=key.get("email"),
            email_confirmed=key.get("email_confirmed"),
            has_password=key.get("has_password"),
            username=key.get("username"),
            first_name=key.get("first_name"),
            last_name=key.get("last_name"),
            picture_url=key.get("picture_url"),
            locked=key.get("locked"),
            enabled=key.get("enabled"),
            mfa_enabled=key.get("mfa_enabled"),
            can_create_orgs=key.get("can_create_orgs"),
            created_at=key.get("created_at"),
            last_active_at=key.get("last_active_at"),
            org_id_to_org_info=_to_org_member_info(key.get("org_id_to_org_info")),
            legacy_user_id=key.get("legacy_user_id"),
            impersonator_user_id=key.get("impersonator_user_id"),
            metadata=key.get("metadata"),
            properties=key.get("properties"),
        )
        for key in json_response.get("users")
    ]

    return UsersPagedResponse(
        users=users,
        total_users=json_response.get("total_users"),
        current_page=json_response.get("current_page"),
        page_size=json_response.get("page_size"),
        has_more_results=json_response.get("has_more_results"),
    )
    
    
def _deserialize_mfa_setup(mfa_data: Optional[Dict[str, Any]]) -> Optional[Union[MfaTotpType, MfaPhoneType]]:
    if mfa_data is None:
        return None
    
    mfa_type = mfa_data.get("type")
    
    if mfa_type == MfaTotpType.type:
        return MfaTotpType(type=mfa_type)
    elif mfa_type == MfaPhoneType.type:
        phone_numbers = [
            MfaPhones(
                mfa_phone_number_suffix=phone["mfa_phone_number_suffix"],
                mfa_phone_id=phone["mfa_phone_id"]
            )
            for phone in mfa_data.get("phone_numbers", [])
        ]
        return MfaPhoneType(type=mfa_type, phone_numbers=phone_numbers)
    else:
        raise ValueError(f"Unknown MFA type: {mfa_type}")
    
def _fetch_user_mfa_methods(
    auth_hostname, integration_api_key, user_id
) -> Optional[FetchUserMfaMethodsResponse]:
    response = requests.get(
        url=f"{ENDPOINT_URL}/{user_id}/mfa",
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif response.status_code == 404:
        return None
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching user mfa methods")

    json_response = response.json()
    mfa_setup = _deserialize_mfa_setup(json_response.get("mfa_setup"))

    return FetchUserMfaMethodsResponse(mfa_setup=mfa_setup)
    
async def _fetch_user_mfa_methods_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id,
) -> Optional[FetchUserMfaMethodsResponse]:
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.get(
        url=f"{ENDPOINT_URL}/{user_id}/mfa",
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise BadRequestException(response.text)
    elif response.status_code == 404:
        return None

    response.raise_for_status()

    json_response = response.json()
    mfa_setup = _deserialize_mfa_setup(json_response.get("mfa_setup"))

    return FetchUserMfaMethodsResponse(mfa_setup=mfa_setup)


####################
#       POST     #
####################
def _create_user(
    auth_hostname,
    integration_api_key,
    email,
    email_confirmed,
    send_email_to_confirm_email_address,
    ask_user_to_update_password_on_login,
    password,
    username,
    first_name,
    last_name,
    properties,
    ignore_domain_restrictions,
) -> CreatedUser:
    url = f"{ENDPOINT_URL}/"
    json = {
        "email": email,
        "email_confirmed": email_confirmed,
        "send_email_to_confirm_email_address": send_email_to_confirm_email_address,
        "ask_user_to_update_password_on_login": ask_user_to_update_password_on_login,
    }
    if password is not None:
        json["password"] = password
    if username is not None:
        json["username"] = username
    if first_name is not None:
        json["first_name"] = first_name
    if last_name is not None:
        json["last_name"] = last_name
    if properties is not None:
        json["properties"] = properties
    if ignore_domain_restrictions is not None:
        json["ignore_domain_restrictions"] = ignore_domain_restrictions

    response = requests.post(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise CreateUserException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when creating user")

    json_response = response.json()
    return CreatedUser(user_id=json_response.get("user_id"))

async def _create_user_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    email,
    email_confirmed,
    send_email_to_confirm_email_address,
    ask_user_to_update_password_on_login,
    password,
    username,
    first_name,
    last_name,
    properties,
    ignore_domain_restrictions,
) -> CreatedUser:
    url = f"{ENDPOINT_URL}/"
    json_body = {
        "email": email,
        "email_confirmed": email_confirmed,
        "send_email_to_confirm_email_address": send_email_to_confirm_email_address,
        "ask_user_to_update_password_on_login": ask_user_to_update_password_on_login,
    }
    # Add optional fields if they are provided
    if password is not None:
        json_body["password"] = password
    if username is not None:
        json_body["username"] = username
    if first_name is not None:
        json_body["first_name"] = first_name
    if last_name is not None:
        json_body["last_name"] = last_name
    if properties is not None:
        json_body["properties"] = properties
    if ignore_domain_restrictions is not None:
        json_body["ignore_domain_restrictions"] = ignore_domain_restrictions

    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.post(
        url,
        json=json_body,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise CreateUserException(response.json())

    response.raise_for_status()

    json_response = response.json()
    return CreatedUser(user_id=json_response.get("user_id"))


def _disable_user(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/disable"

    response = requests.post(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when disabling user")

    return True

async def _disable_user_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/disable"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.post(
        url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


def _enable_user(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/enable"

    response = requests.post(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when enabling user")

    return True

async def _enable_user_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/enable"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.post(
        url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


def _disable_user_2fa(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/disable_2fa"

    response = requests.post(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when enabling user")

    return True

async def _disable_user_2fa_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/disable_2fa"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.post(
        url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


def _invite_user_to_org(
    auth_hostname, integration_api_key, email, org_id, role, additional_roles=[]
) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = BACKEND_API_BASE_URL + "/api/backend/v1/invite_user"
    json = {
        "email": email,
        "org_id": org_id,
        "role": role,
        "additional_roles": additional_roles,
    }
    response = requests.post(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        try:
            response_json = response.json()
            raise InviteUserToOrgException(response_json)
        except requests.exceptions.JSONDecodeError:
            raise ValueError("Bad request: " + response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating metadata")

    return True

async def _invite_user_to_org_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    email,
    org_id,
    role,
    additional_roles=[],
) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = BACKEND_API_BASE_URL + "/api/backend/v1/invite_user"
    json_body = {
        "email": email,
        "org_id": org_id,
        "role": role,
        "additional_roles": additional_roles if additional_roles else [], # Ensure it's a list
    }
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.post(
        url,
        json=json_body,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        try:
            raise InviteUserToOrgException(response.json())
        except Exception:
             raise BadRequestException(response.text)
    elif response.status_code == 404:
        return False 

    response.raise_for_status()
    return True

def _invite_user_to_org_by_user_id(
    auth_hostname, integration_api_key, user_id, org_id, role, additional_roles=[]
) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = BACKEND_API_BASE_URL + "/api/backend/v1/invite_user_by_id"
    json = {
        "user_id": user_id,
        "org_id": org_id,
        "role": role,
        "additional_roles": additional_roles,
    }
    response = requests.post(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        try:
            response_json = response.json()
            raise InviteUserToOrgException(response_json)
        except requests.exceptions.JSONDecodeError:
            raise ValueError("Bad request: " + response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating metadata")

    return True

async def _invite_user_to_org_by_user_id_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id,
    org_id,
    role,
    additional_roles=[],
) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = BACKEND_API_BASE_URL + "/api/backend/v1/invite_user_by_id"
    json_body = {
        "user_id": user_id,
        "org_id": org_id,
        "role": role,
        "additional_roles": additional_roles if additional_roles else [], # Ensure it's a list
    }
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.post(
        url,
        json=json_body,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        try:
            raise InviteUserToOrgException(response.json())
        except Exception:
             raise BadRequestException(response.text)
    elif response.status_code == 404:
        return False 

    response.raise_for_status()
    return True


def _resend_email_confirmation(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = BACKEND_API_BASE_URL + "/api/backend/v1/resend_email_confirmation"

    json = {
        "user_id": user_id,
    }

    response = requests.post(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif response.status_code == 429:
        try:
            # Check if this is specifically an email-send rate limit error
            error_message = response.json()["user_facing_error"]
            raise RateLimitedException(error_message)
        except requests.exceptions.JSONDecodeError:
            raise RateLimitedException(response.text)
    elif response.status_code == 400:
        if response.json().get("user_facing_error"):
            raise ValueError(response.json().get("user_facing_error"))
        else:
            raise RuntimeError("Unknown error when resending email confirmation")
    elif not response.ok:
        raise RuntimeError("Unknown error when resending email confirmation")

    return True

async def _resend_email_confirmation_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = BACKEND_API_BASE_URL + "/api/backend/v1/resend_email_confirmation"
    json_body = {"user_id": user_id}
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.post(
        url,
        json=json_body,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif response.status_code == 429:
        try:
            error_message = response.json().get("user_facing_error", response.text)
            raise RateLimitedException(error_message)
        except httpx.ResponseNotRead:
            raise RateLimitedException(response.text)
        except Exception:
            raise RateLimitedException(response.text)
    elif response.status_code == 400:
        try:
            error_message = response.json().get("user_facing_error")
            if error_message:
                raise ValueError(error_message)
            else:
                raise BadRequestException(response.text)
        except httpx.ResponseNotRead:
             raise BadRequestException(response.text)
        except Exception:
            raise BadRequestException(response.text)


    response.raise_for_status()
    return True


def _logout_all_user_sessions(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/logout_all_sessions"

    response = requests.post(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when logging out all user sessions")

    return True

async def _logout_all_user_sessions_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/logout_all_sessions"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.post(
        url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


####################
#     PATCH/PUT    #
####################
def _update_user_metadata(
    auth_hostname,
    integration_api_key,
    user_id,
    username=None,
    first_name=None,
    last_name=None,
    metadata=None,
    properties=None,
    picture_url=None,
    update_password_required=None,
    legacy_user_id=None,
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}"
    json = {}
    if username is not None:
        json["username"] = username
    if first_name is not None:
        json["first_name"] = first_name
    if last_name is not None:
        json["last_name"] = last_name
    if metadata is not None:
        json["metadata"] = metadata
    if properties is not None:
        json["properties"] = properties
    if picture_url is not None:
        json["picture_url"] = picture_url
    if update_password_required is not None:
        json["update_password_required"] = update_password_required
    if legacy_user_id is not None:
        json["legacy_user_id"] = legacy_user_id

    response = requests.put(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise UpdateUserMetadataException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating metadata")

    return True

async def _update_user_metadata_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id,
    username=None,
    first_name=None,
    last_name=None,
    metadata=None,
    properties=None,
    picture_url=None,
    update_password_required=None,
    legacy_user_id=None,
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}"
    json_body = {}
    if username is not None:
        json_body["username"] = username
    if first_name is not None:
        json_body["first_name"] = first_name
    if last_name is not None:
        json_body["last_name"] = last_name
    if metadata is not None:
        json_body["metadata"] = metadata
    if properties is not None:
        json_body["properties"] = properties
    if picture_url is not None:
        json_body["picture_url"] = picture_url
    if update_password_required is not None:
        json_body["update_password_required"] = update_password_required
    if legacy_user_id is not None:
        json_body["legacy_user_id"] = legacy_user_id

    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.put(
        url,
        json=json_body,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise UpdateUserMetadataException(response.json())
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


def _update_user_password(
    auth_hostname,
    integration_api_key,
    user_id,
    password,
    ask_user_to_update_password_on_login,
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/password"
    json = {"password": password}
    if ask_user_to_update_password_on_login is not None:
        json["ask_user_to_update_password_on_login"] = (
            ask_user_to_update_password_on_login
        )

    response = requests.put(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise UpdateUserPasswordException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating password")

    return True

async def _update_user_password_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id,
    password,
    ask_user_to_update_password_on_login,
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/password"
    json_body = {"password": password}
    if ask_user_to_update_password_on_login is not None:
        json_body["ask_user_to_update_password_on_login"] = (
            ask_user_to_update_password_on_login
        )

    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.put(
        url,
        json=json_body,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise UpdateUserPasswordException(response.json())
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


def _clear_user_password(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/clear_password"

    response = requests.put(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise UpdateUserEmailException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating user email")

    return True

async def _clear_user_password_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/clear_password"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.put(
        url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise UpdateUserEmailException(response.json())
    elif response.status_code == 404:
        return False 

    response.raise_for_status()
    return True


def _update_user_email(
    auth_hostname, integration_api_key, user_id, new_email, require_email_confirmation
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/email"
    json = {
        "new_email": new_email,
        "require_email_confirmation": require_email_confirmation,
    }

    response = requests.put(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise UpdateUserEmailException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating user email")

    return True

async def _update_user_email_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id,
    new_email,
    require_email_confirmation
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/email"
    json_body = {
        "new_email": new_email,
        "require_email_confirmation": require_email_confirmation,
    }
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.put(
        url,
        json=json_body,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise UpdateUserEmailException(response.json())
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


def _enable_user_can_create_orgs(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/can_create_orgs/enable"

    response = requests.put(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when enabling can_create_orgs")

    return True

async def _enable_user_can_create_orgs_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/can_create_orgs/enable"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.put(
        url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


def _disable_user_can_create_orgs(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/can_create_orgs/disable"

    response = requests.put(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when disabling can_create_orgs")

    return True

async def _disable_user_can_create_orgs_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}/can_create_orgs/disable"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.put(
        url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


####################
#       DELETE     #
####################
def _delete_user(auth_hostname, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}"

    response = requests.delete(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when deleting user")

    return True

async def _delete_user_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    user_id
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = f"{ENDPOINT_URL}/{user_id}"
    headers = _get_async_headers(auth_hostname, integration_api_key)

    response = await httpx_client.delete(
        url,
        headers=headers,
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 404:
        return False

    response.raise_for_status()
    return True


####################
#       HELPERS    #
####################


def _validate_personal_api_key(
    auth_hostname, integration_api_key, api_key_token
) -> PersonalApiKeyValidation:
    api_key_validation = _validate_api_key(
        auth_hostname, integration_api_key, api_key_token
    )
    if not api_key_validation.user or api_key_validation.org:
        raise EndUserApiKeyException({"api_key_token": ["Not a personal API Key"]})
    return PersonalApiKeyValidation(
        user=api_key_validation.user,
        metadata=api_key_validation.metadata,
    )

async def _validate_personal_api_key_async(
    httpx_client: httpx.AsyncClient,
    auth_hostname,
    integration_api_key,
    api_key_token
) -> PersonalApiKeyValidation:
    api_key_validation = await _validate_api_key_async(
        httpx_client=httpx_client,
        auth_hostname=auth_hostname,
        integration_api_key=integration_api_key,
        api_key_token=api_key_token
    )
    if not api_key_validation.user or api_key_validation.org:
        raise EndUserApiKeyException({"api_key_token": ["Not a personal API Key"]})

    return PersonalApiKeyValidation(
        user=api_key_validation.user,
        metadata=api_key_validation.metadata,
    )



import inspect

from propelauth_py import init_base_auth
from propelauth_py.api.access_token import _create_access_token
from propelauth_py.api.end_user_api_keys import (
    _create_api_key,
    _delete_api_key,
    _fetch_api_key,
    _fetch_archived_api_keys,
    _fetch_current_api_keys,
    _update_api_key,
    _validate_api_key,
    _fetch_api_key_usage,
    _import_api_key,
    _validate_imported_api_key
)
from propelauth_py.api.step_up_mfa.send_sms_mfa_code import _send_sms_mfa_code
from propelauth_py.api.step_up_mfa.verify_grant import _verify_step_up_grant
from propelauth_py.api.step_up_mfa.verify_sms_challenge import _verify_sms_challenge
from propelauth_py.api.step_up_mfa.verify_totp_challenge import _verify_step_up_totp_challenge
from propelauth_py.api.magic_link import _create_magic_link
from propelauth_py.api.migrate_user import _migrate_user_from_external_source, _migrate_user_password
from propelauth_py.api.org import (
    _add_user_to_org,
    _allow_org_to_setup_saml_connection,
    _change_user_role_in_org,
    _create_org,
    _create_org_saml_connection_link,
    _delete_org,
    _disallow_org_to_setup_saml_connection,
    _fetch_custom_role_mappings,
    _fetch_org,
    _fetch_org_by_query,
    _fetch_pending_invites,
    _remove_user_from_org,
    _revoke_pending_org_invite,
    _subscribe_org_to_role_mapping,
    _update_org_metadata,
    _validate_org_api_key,
    _fetch_saml_sp_metadata,
    _set_saml_idp_metadata,
    _saml_go_live,
    _delete_saml_connection,
)
from propelauth_py.api.user import (
    _clear_user_password,
    _create_user,
    _delete_user,
    _disable_user,
    _disable_user_2fa,
    _disable_user_can_create_orgs,
    _enable_user,
    _enable_user_can_create_orgs,
    _fetch_batch_user_metadata_by_emails,
    _fetch_batch_user_metadata_by_user_ids,
    _fetch_batch_user_metadata_by_usernames,
    _fetch_user_metadata_by_email,
    _fetch_user_metadata_by_user_id,
    _fetch_user_metadata_by_username,
    _fetch_user_signup_query_params_by_user_id,
    _fetch_users_by_query,
    _fetch_users_in_org,
    _invite_user_to_org,
    _invite_user_to_org_by_user_id,
    _logout_all_user_sessions,
    _resend_email_confirmation,
    _update_user_email,
    _update_user_metadata,
    _update_user_password,
    _validate_personal_api_key,
    _fetch_user_mfa_methods,
)
from propelauth_py.api.employee import (
    _fetch_employee_by_id,
)
from tests.conftest import TEST_BASE_AUTH_URL, mock_api_and_init_auth, rsa_keys

IMPORTED_FUNCTIONS = [
    _fetch_user_metadata_by_user_id,
    _fetch_user_metadata_by_email,
    _fetch_user_metadata_by_username,
    _fetch_batch_user_metadata_by_user_ids,
    _fetch_batch_user_metadata_by_emails,
    _fetch_batch_user_metadata_by_usernames,
    _fetch_user_signup_query_params_by_user_id,
    _fetch_users_by_query,
    _fetch_users_in_org,
    _create_user,
    _update_user_email,
    _update_user_metadata,
    _delete_user,
    _disable_user,
    _enable_user,
    _update_user_password,
    _clear_user_password,
    _disable_user_2fa,
    _enable_user_can_create_orgs,
    _disable_user_can_create_orgs,
    _validate_personal_api_key,
    _fetch_org,
    _fetch_org_by_query,
    _fetch_pending_invites,
    _fetch_custom_role_mappings,
    _create_org,
    _remove_user_from_org,
    _update_org_metadata,
    _subscribe_org_to_role_mapping,
    _add_user_to_org,
    _allow_org_to_setup_saml_connection,
    _disallow_org_to_setup_saml_connection,
    _validate_org_api_key,
    _create_magic_link,
    _create_access_token,
    _migrate_user_from_external_source,
    _migrate_user_password,
    _fetch_api_key,
    _fetch_current_api_keys,
    _fetch_archived_api_keys,
    _create_api_key,
    _update_api_key,
    _delete_api_key,
    _validate_api_key,
    _change_user_role_in_org,
    _delete_org,
    _invite_user_to_org,
    _invite_user_to_org_by_user_id,
    _resend_email_confirmation,
    _logout_all_user_sessions,
    _revoke_pending_org_invite,
    _create_org_saml_connection_link,
    _fetch_saml_sp_metadata,
    _set_saml_idp_metadata,
    _saml_go_live,
    _delete_saml_connection,
    _fetch_employee_by_id,
    _fetch_api_key_usage,
    _import_api_key,
    _validate_imported_api_key,
    _fetch_user_mfa_methods,
    _verify_step_up_totp_challenge,
    _send_sms_mfa_code,
    _verify_step_up_grant,
    _verify_sms_challenge
]


def test_all_functions_imported_in_init_base_auth():
    """
    Test that all API functions are imported in init_base_auth().
    """
    init_base_auth_nested_functions = [
        name.co_name
        for name in init_base_auth.__code__.co_consts
        if inspect.iscode(name)
    ]

    imported_functions_without_underscores = [
        func.__name__[1:] for func in IMPORTED_FUNCTIONS
    ]
    missing_functions = [
        func_name
        for func_name in init_base_auth_nested_functions
        if func_name not in imported_functions_without_underscores
    ]

    assert len(missing_functions) == 0, f"Missing functions: {missing_functions}"


def test_auth_tuple_contains_all_expected_functions(rsa_keys):
    """
    Test that the Auth tuple contains all expected functions.
    """
    mock_auth_class = mock_api_and_init_auth(
        TEST_BASE_AUTH_URL, 200, {"verifier_key_pem": rsa_keys.public_pem}
    )
    auth_methods = [
        func for func in dir(mock_auth_class)
        if callable(getattr(mock_auth_class, func)) and not func.startswith("_")
    ]
    imported_functions_without_underscores = [
        func.__name__[1:] for func in IMPORTED_FUNCTIONS
    ]
    missing_functions = [
        func_name
        for func_name in imported_functions_without_underscores
        if func_name not in auth_methods
    ]

    assert len(missing_functions) == 0, f"Missing functions: {missing_functions}"



import pytest

from propelauth_py.errors import ForbiddenException
from tests.auth_helpers import create_access_token, orgs_to_org_id_map, random_org, random_user_id

def test_validate_minimum_org_role_and_get_org(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin")
    org["inherited_user_roles_plus_current_role"] = ["Admin", "Member"]
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    try:
        auth.validate_access_token_and_get_user_with_org_by_minimum_role("Bearer " + access_token, org["org_id"], "Admin")
    except Exception as exc:
        assert False, f"'validate_access_token_and_get_user_with_org_by_minimum_role with Admin' raised an exception {exc}"

    try:
        auth.validate_access_token_and_get_user_with_org_by_minimum_role("Bearer " + access_token, org["org_id"], "Member")
    except Exception as exc:
        assert False, f"'validate_access_token_and_get_user_with_org_by_minimum_role with Member' raised an exception {exc}"


def test_validate_exact_org_role_and_get_org(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin")
    org["inherited_user_roles_plus_current_role"] = ["Admin", "Member"]
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    try:
        auth.validate_access_token_and_get_user_with_org_by_exact_role("Bearer " + access_token, org["org_id"], "Admin")
    except Exception as exc:
        assert False, f"'validate_access_token_and_get_user_with_org_by_exact_role with Admin' raised an exception {exc}"

    with pytest.raises(ForbiddenException):
        auth.validate_access_token_and_get_user_with_org_by_exact_role("Bearer " + access_token, org["org_id"], "Member")

def test_validate_permission_and_get_org(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin")
    org["user_permissions"] = ["read", "write"]
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    try:
        auth.validate_access_token_and_get_user_with_org_by_permission("Bearer " + access_token, org["org_id"], "read")
    except Exception as exc:
        assert False, f"'validate_access_token_and_get_user_with_org_by_permission with read' raised an exception {exc}"

    with pytest.raises(ForbiddenException):
        auth.validate_access_token_and_get_user_with_org_by_permission("Bearer " + access_token, org["org_id"], "delete")


def test_validate_all_permissions_and_get_org(auth, rsa_keys):
    user_id = random_user_id()
    org = random_org("Admin")
    org["user_permissions"] = ["read", "write"]
    org_id_to_org_member_info = orgs_to_org_id_map([org])
    access_token = create_access_token({
        "user_id": user_id,
        "org_id_to_org_member_info": org_id_to_org_member_info
    }, rsa_keys.private_pem)

    try:
        auth.validate_access_token_and_get_user_with_org_by_all_permissions("Bearer " + access_token, org["org_id"], ["read", "write"])
    except Exception as exc:
        assert False, f"'validate_access_token_and_get_user_with_org_by_all_permissions with read, write' raised an exception {exc}"

    with pytest.raises(ForbiddenException):
        auth.validate_access_token_and_get_user_with_org_by_all_permissions("Bearer " + access_token, org["org_id"], ["read", "write", "delete"])

